#include "CorrespondencePointMatcher.h"

#include <vector>
#include <memory>
#include <algorithm>

#include <unordered_set>
#include <chrono>

TArray<TArray<FVector>> CorrespondencePointMatcher::matchOnlyGridCenters(TArray<GridDetectionResult> results) {
	TArray<TArray<FVector>> result;

	for (GridDetectionResult gdr : results) {
		TArray<FVector> singleResult;

		if (gdr.gridFound) {
			singleResult.Add(gdr.gridCenter);
		}
		else
		{
			singleResult.Add(FVector(NAN));
		}

		result.Add(singleResult);
	}

	return result;
}

TArray<TArray<FVector>> CorrespondencePointMatcher::matchAllGridHoles_ByRoughRegistration(TArray<FTransform> roughRegistration, TArray<GridDetectionResult> results, int referenceID) {
	if (roughRegistration.Num() != results.Num()) {
		UE_LOG(LogTemp, Warning, TEXT("matchAllGridHoles_ByRoughRegistration: Both input arrays must have the same size!"));
		return TArray<TArray<FVector>>();
	}

	if (roughRegistration.Num() < 2) {
		UE_LOG(LogTemp, Warning, TEXT("matchAllGridHoles_ByRoughRegistration: There must be at least 2 array elements for registration!"));
		return TArray<TArray<FVector>>();
	}

	FVector referenceNormal = results[referenceID].normal;

	TArray<TArray<FVector>> result;

	for (int i = 0; i < results.Num(); ++i) {
		GridDetectionResult gdr = results[i];
		FTransform transform = roughRegistration[i];
		FVector normalTransformed = transform.TransformVector(gdr.normal);

		bool otherSideThanReference = FVector::DotProduct(referenceNormal,normalTransformed) < 0;

		TArray<FVector> singleResult;
		for (int x = 0; x < 5; ++x) {
			for (int y = 0; y < 5; ++y) {
				float expX = (x - 2) * 0.08f;
				float expY = (y - 2) * 0.08f;

				if (otherSideThanReference) {
					expY = (2 - y) * 0.08f;
				}

				float minDistSqr = 999999;
				FVector minPoint;

				for (FVector holeCenter : gdr.centersOfFoundHoles) {
					FVector v = holeCenter - gdr.gridCenter;

					float projX = FVector::DotProduct(gdr.axis1, v);
					float projY = FVector::DotProduct(gdr.axis2, v);

					float dX = expX - projX;
					float dY = expY - projY;

					float distSqr = dX * dX + dY * dY;

					if (distSqr < minDistSqr) {
						minDistSqr = distSqr;
						minPoint = holeCenter;
					}
				}

				if (minDistSqr < 0.03 * 0.03f) {
					singleResult.Add(minPoint);
				}
				else
				{
					singleResult.Add(FVector(NAN));
				}
			}
		}

		result.Add(singleResult);
	}

	return result;
}


TArray<TArray<FVector>> CorrespondencePointMatcher::matchGridCenterAndXAxis(TArray<GridDetectionResult> results) {
	TArray<TArray<FVector>> result;

	for (GridDetectionResult gdr : results) {
		TArray<FVector> singleResult;

		if (gdr.gridFound) {
			singleResult.Add(gdr.gridCenter);
			singleResult.Add(gdr.gridCenter + gdr.axis1 * 0.2f);
		}
		else 
		{
			singleResult.Add(FVector(NAN));
			singleResult.Add(FVector(NAN));
		}

		result.Add(singleResult);
	}

	return result;
}

/*

std::vector<std::pair<pcl::PointXYZ, pcl::PointXYZ>> matchForRoughRegistration(GridRecognitionResult r1, GridRecognitionResult r2) {
	std::vector<std::pair<pcl::PointXYZ, pcl::PointXYZ>> funcResult;

	std::pair<pcl::PointXYZ, pcl::PointXYZ> pair(
		pcl::PointXYZ(r1.gridCenter.x(), r1.gridCenter.y(), r1.gridCenter.z()),
		pcl::PointXYZ(r2.gridCenter.x(), r2.gridCenter.y(), r2.gridCenter.z())
	);

	std::pair<pcl::PointXYZ, pcl::PointXYZ> pair2(
		pcl::PointXYZ(r1.gridCenter.x() + r1.axis1.x() * 0.1f, r1.gridCenter.y() + r1.axis1.y() * 0.1f, r1.gridCenter.z() + r1.axis1.z() * 0.1f),
		pcl::PointXYZ(r2.gridCenter.x() + r2.axis1.x() * 0.1f, r2.gridCenter.y() + r2.axis1.y() * 0.1f, r2.gridCenter.z() + r2.axis1.z() * 0.1f)
	);

	funcResult.push_back(pair);
	funcResult.push_back(pair2);

	return funcResult;
}

*/

TArray<TArray<FVector>> CorrespondencePointMatcher::matchAllGridHoles_AssumingSameSides(TArray<GridDetectionResult> results) {
	TArray<TArray<FVector>> result;

	for (GridDetectionResult gdr : results) {
		TArray<FVector> singleResult;

		for (int x = 0; x < 5; ++x) {
			for (int y = 0; y < 5; ++y) {
				float expX = (x - 2) * 0.08f;
				float expY = (y - 2) * 0.08f;

				float minDistSqr = 999999;
				FVector minPoint;

				for (FVector holeCenter : gdr.centersOfFoundHoles) {
					FVector v = holeCenter - gdr.gridCenter;

					float projX = FVector::DotProduct(gdr.axis1, v);
					float projY = FVector::DotProduct(gdr.axis2, v);

					float dX = expX - projX;
					float dY = expY - projY;

					float distSqr = dX * dX + dY * dY;

					if (distSqr < minDistSqr) {
						minDistSqr = distSqr;
						minPoint = holeCenter;
					}
				}

				if (minDistSqr < 0.03 * 0.03f) {
					singleResult.Add(minPoint);
				}
				else
				{
					singleResult.Add(FVector(NAN));
				}
			}
		}

		result.Add(singleResult);
	}

	return result;
}

MotionControllerResult CorrespondencePointMatcher::findMotionControllerRelativeToGrid(GridDetectionResult gdr, MotionControllerAssumptions assumptions) {
	// If no grid was visible, motion controller cannot be found:
	if (!gdr.gridFound)
		return MotionControllerResult();

	MotionControllerResult result;
	result.found = true;

	FVector leftOffset = assumptions.swapControllers ? assumptions.rightOffsetFromGridMid : assumptions.leftOffsetFromGridMid;
	FVector rightOffset = assumptions.swapControllers ? assumptions.leftOffsetFromGridMid : assumptions.rightOffsetFromGridMid;

	FTransform transform(gdr.axis1, gdr.axis2, gdr.normal, gdr.gridCenter);
	result.leftControllerPosition = transform.TransformPosition(leftOffset);
	result.rightControllerPosition = transform.TransformPosition(rightOffset);

	return result;
}

TArray<TArray<FVector>> CorrespondencePointMatcher::matchControllerAndGridDetectionResult(FVector leftControllerWS, FVector rightControllerWS, MotionControllerResult mcr) {
	TArray<TArray<FVector>> result;

	result.Add(TArray<FVector>());
	result.Add(TArray<FVector>());

	if (mcr.found) {
		result[0].Add(leftControllerWS);
		result[0].Add(rightControllerWS);

		result[1].Add(mcr.leftControllerPosition);
		result[1].Add(mcr.rightControllerPosition);
	}

	return result;
}