//  2021, CGVR (https://cgvr.informatik.uni-bremen.de/)
// 
// LICENSE:
// This original work is copyright by University of Bremen.
// Any software of this work is covered by the European Union Public Licence v1.2.To view a copy of this
// license, visit eur-lex.europa.eu. If you use software to produce a publication, then you must give credit 
// and put a reference in your publication. If you would like to use our software in proprietary software, 
// you can obtain an exception from the above license(aka. dual licensing).
// Please contact zach@cs.uni-bremen.de.
//
// Code by Andre Mhlenbrock (muehlenb@uni-bremen.de)
//
// This class implements the algorithm for the grid detection presented in the paper 
// "Fast and Robust 3D Grid Target Detection for Registration and Calibration of Depth-Only Sensors"
//
// A few hints, because the code of this algorithm is quite extensive, not quite trivial and some weeks 
// of work went into it:
//
// - With the current parameters, a grid of the size 44x44 with 6x6 bars and 25 holes in the size of 4cm x 4cm
//   is recognized. The parameters and the algorithm IS TESTED with the Kinect v2 and Azure Kinect - they work
//   pretty well.
//
// - AT THE BOTTOM, you find the implementation of the detectGrid-Method. Outgoing from this method, all 
// above declared methods will be called. This is also the best documented method. If you want to understand
// the program code, it could be helpful to start there.
// 
// - This algorithm detects grids of a fixed size on depth data, which is a simple float* array. For speed 
// optimization, it is assumed that the point cloud is in its arrangement of the original depth image. 
// That is, points in the point cloud with consecutive indices were adjacent to each other in the original 
// depth image (unless they are image edge points).
//
// - The Scanline3D data structure, which is used multiple times in the algorithm, represents a scanline in
// the original assumed depth image and also implements the union-find structure so that scanlines can have
// other scanlines as parents and can form regions (in the depth image) via the resulting tree structure. 
// Furthermore, this Scanline3D data structure only works with indices of the passed point cloud, not with 
// the point data itself.
//
// - This algorithm expects METERS for the values of the points. If you work with cm or mm, scale them or
// change the parameters.
// 
// If you have questions, feel free to ask them to me: Andre Mhlenbrock (muehlenb@uni-bremen.de)
// 

// The maximum square distance of two consecutive points to get merged into one scanline:
#define SCANLINE_CREATION_MAX_SQR_POINT_DISTANCE 0.0025f 

// Minimal distance of a scanline in front of or behind a hole scanline (e.g. created by a single grid border):
#define SCANLINE_MIN_HOLE_SURROUNDING_SIZE 0.035f // 3.5cm because a border has a size of 4cm minus a little delta.

// Maximal distance of a scanline in front of or behind a hole scanline (e.g. created by multiple grid borders by 
// scaning over a slightly rotated grid):
#define SCANLINE_MAX_HOLE_SURROUNDING_SIZE 0.35f // 35cm, because while scanning over a slightly rotated grid (in 2D)
												 // there can be scanlines which run over multiple grid borders which
												 // are horizontally and vertically ordered.

#define SCANLINE_MAX_HOLE_SIZE 0.045f // consider a hole length of maximum sqrt(4cm*4cm + 4cm*4cm) along a scanline - 
									  // perhaps this value should be a bit bigger but well, it works quite good...

// Maximal square distance of the midpoints of two scanlines to be merged into one segment:
#define SCANLINE_MERGING_MAX_SQR_MIDPOINT_DISTANCE 0.0144f

// Describes how much each the x- and y- eigenvalue must be bigger than the z-eigenvalue
// to be able to be a grid candidate (a grid is flat, so the z-value must be very small):
#define PCA_MIN_XY_TO_Z_RATIO 10

// Don't consider segments which are to small:
#define PCA_MIN_POINT_SIZE 20

// Maximum difference of the x-eigenvalue and y-eigenvalue as a factor (a grid will
// produce relative equal values in x- and y- direction).
#define PCA_MAX_X_TO_Y_RATION 2

// Maximum squared distance for points to be removed for performance reasons:
#define REDUCE_POINTS_MAX_SQR_DISTANCE 0.0004f

// Radius of the considered point cloud around the points of the grid candidate:
#define FILTER_PC_MAX_SQR_RADIUS 0.16f * 0.16f

// Minimum number of points in the filtered radius to be able to be recognize as a grid:
#define FILTER_PC_MIN_SIZE_OF_GRID_CANDIDATE 100

// Threshold of the first ransac pass to get inliers and work with them (especially to remove the 
// flying pixel effect of some ToF cameras):
#define RANSAC_FIRST_PASS_THRESHOLD 0.015f

// Threshold of the second ransac pass to estimate the plane model parameters a bit more precise:
#define RANSAC_SECOND_PASS_THRESHOLD 0.005f

// Maximum pixels per side of an hole segment to be considered as hole:
#define GRID_MID_PIXELS_MAX_SIDE_LENGTH 20

// Minimum amount of pixels in a hole segment to be considered as hole:
#define GRID_MID_PIXEL_MIN_PIXEL_COUNT 20

// Maximum distance of two grid mid points (not the best naming for the mid point of holes(!)) to 
// apply our filtering / cluster heuristic to filter wrong points:
#define CLUSTER_MID_POINTS_MAX_DISTANCE 0.10f  // 8cm + delta

// Minimum distance of two grid mid points (not the best naming for the mid point of holes(!)) to 
// apply our filtering / cluster heuristic to filter wrong points:
#define CLUSTER_MID_POINTS_MIN_DISTANCE 0.06f // 8cm - delta

// Maximal radial distance to cluster mid points (or better name: "hole mid points") by our
// filtering / cluster heuristic to filter wrong points:
#define CLUSTER_MID_POINTS_MAX_RADIAL_MERGING_DISTANCE 0.03f

// Minimum size of the two biggest clusters (in our filtering / cluster heuristic) to be considered as a grid:
#define CLUSTER_MID_POINTS_MIN_RADIAL_CLUSTER_SIZE 20

// Maximal radius around the total mid point of the grid in which hole mid points are used for a more precise 
// mid point calculation:
#define MID_POINT_PREDICTION_MAXIMAL_RADIUS 0.24f //(16cm is the maximal distance from the mid point to one of the outer
												  // diagonal hole mid points)

// Distance along one axis (x- or y-axis) in which hole points can be:
#define CLUSTER_MID_POINTS_MAXIMAL_AXIS_DISTANCE_TO_MID 0.18f // 2x8cm (8cm is the distance of two hole mid points) + delta

#include "GridDetector.h"

#include <vector>
#include <memory>
#include <algorithm>

#include <unordered_set>
#include <chrono>

THIRD_PARTY_INCLUDES_START
#include <pcl/io/pcd_io.h>
#include <pcl/point_types.h>
#include <pcl/registration/transformation_estimation.h>
#include <pcl/registration/transformation_estimation_svd.h>

#include <pcl/common/common.h>
#include <pcl/common/pca.h>

#include <pcl/sample_consensus/ransac.h>
#include <pcl/sample_consensus/sac_model_plane.h>
THIRD_PARTY_INCLUDES_END

/**
* Represents a scanline in the image coordinates of the original depth image of 
* the point cloud.
*
* Implements the union find structure for efficient merging of scanlines.
*/

struct Scanline3D
	: public std::enable_shared_from_this<Scanline3D>
{
	std::shared_ptr<Scanline3D> parent;
	int startIdx;
	int endIdx;
	int depth;
	int segmentIdx;
	FVector mean;
	FVector dir;
	float simpleLength;

	Scanline3D(int idx)
		: parent(nullptr)
		, startIdx(idx)
		, endIdx(idx)
		, depth(1)
	{}

	Scanline3D(int startIdx, int endIdx)
		: parent(nullptr)
		, startIdx(startIdx)
		, endIdx(endIdx)
		, depth(1)
	{}

	std::shared_ptr<Scanline3D> getRoot() {
		return getRoot(nullptr, nullptr);
	}

	int lineSize() {
		return endIdx - startIdx + 1;
	}

	std::shared_ptr<Scanline3D> getRoot(std::shared_ptr<Scanline3D> previous, std::shared_ptr<Scanline3D> prePrevious) {
		std::shared_ptr<Scanline3D> self = shared_from_this();

		// path compression:
		if (prePrevious != nullptr) {
			prePrevious->parent = self;
		}

		if (self->parent == nullptr)
			return self;

		return parent->getRoot(self, previous);
	}

	void Union(std::shared_ptr<Scanline3D> node) {
		std::shared_ptr<Scanline3D> aRoot = getRoot(nullptr, nullptr);
		std::shared_ptr<Scanline3D> bRoot = node->getRoot(nullptr, nullptr);

		if (aRoot == bRoot)
			return;

		// attach tree with smaller depth to the tree with higher depth:
		if (aRoot->depth < bRoot->depth) {
			aRoot->parent = bRoot;
			bRoot->depth += aRoot->depth;
		}
		else {
			bRoot->parent = aRoot;
			aRoot->depth += bRoot->depth;
		}
	}
};

/**
* Creates scanlines in the assumed depth image of the given point cloud. Points, those distance is 
* less than SCANLINE_CREATION_MAX_SQR_POINT_DISTANCE are merged in one scanline.
*/

std::vector<std::shared_ptr<Scanline3D>> filterConnectedScanlines(float* input, unsigned int imgWidth, unsigned int imgHeight) {
	std::vector<std::shared_ptr<Scanline3D>> scanlines;

	int pcSize = imgWidth * imgHeight;
	std::shared_ptr<Scanline3D> currentLine = nullptr;

	for (int i = 0; i < pcSize; ++i) {
		// No scanline can leave it's row:
		if (i % imgWidth == 0) {
			if (currentLine != nullptr) {
				scanlines.push_back(currentLine);
			}
			currentLine = std::shared_ptr<Scanline3D>(new Scanline3D(i));
		}
		else {
			// Check whether the distance of the current point to the last point is low enough:
			float* lastPtr = &input[(i - 1)*3];
			FVector last(lastPtr[0], lastPtr[1], lastPtr[2]);

			float* currentPtr = &input[i*3];
			FVector current(currentPtr[0], currentPtr[1], currentPtr[2]);

			float distance = (last - current).SizeSquared();
			
			if (distance < SCANLINE_CREATION_MAX_SQR_POINT_DISTANCE) {
				currentLine->endIdx = i;
			}
			else {
				scanlines.push_back(currentLine);
				currentLine = std::shared_ptr<Scanline3D>(new Scanline3D(i));
			}

		}
	}

	return scanlines;
}

/**
* Filters holes in scanlines (holes are surrounding by two scanlines of the defined size (SCANLINE_MIN_HOLE_SURROUNDING_SIZE, SCANLINE_MAX_HOLE_SURROUNDING_SIZE).
*/

std::vector<std::shared_ptr<Scanline3D>> filterHolesInStraigtLines(float* pcData, unsigned int imageWidth, std::vector<std::shared_ptr<Scanline3D>> input) {
	std::vector<std::shared_ptr<Scanline3D>> result;

	int currentLine = 0;
	std::vector<std::shared_ptr<Scanline3D>> currentLines;

	for (std::shared_ptr<Scanline3D> line : input) {
		if (line->startIdx / imageWidth != currentLine) {
			currentLine = line->startIdx / imageWidth;
			currentLines.clear();
		}

		float* startPtr = &pcData[line->startIdx * 3];
		FVector start(startPtr[0], startPtr[1], startPtr[2]);

		float* endPtr = &pcData[line->endIdx * 3];
		FVector end(endPtr[0], endPtr[1], endPtr[2]);

		float length = (start - end).Size();

		if (length > SCANLINE_MIN_HOLE_SURROUNDING_SIZE && length < SCANLINE_MAX_HOLE_SURROUNDING_SIZE) {
			if (currentLines.size() > 1) {
				std::shared_ptr<Scanline3D> lastLine = currentLines[currentLines.size() - 1];

				float* lastEndPtr = &pcData[lastLine->endIdx * 3];
				FVector lastEnd(lastEndPtr[0], lastEndPtr[1], lastEndPtr[2]);

				if ((lastEnd - start).Size() < SCANLINE_MAX_HOLE_SIZE) {
					std::shared_ptr<Scanline3D> holeLine = std::shared_ptr<Scanline3D>(new Scanline3D(lastLine->endIdx, line->startIdx));
					result.push_back(holeLine);
				}
			}
			currentLines.push_back(line);
		}
	}

	return result;
}

/*
* Merge neightbour scanlines those mid points have a sqr distance less than SCANLINE_MERGING_MAX_SQR_MIDPOINT_DISTANCE.
* 
* Currently not optimized O(n).
*/

void mergeNeighbours(float* pcData, std::vector<std::shared_ptr<Scanline3D>>& input) {
	int inputSize = input.size();

	for (int aIdx = 0; aIdx < inputSize; ++aIdx) {
		std::shared_ptr<Scanline3D> a = input[aIdx];

		for (int bIdx = aIdx + 1; bIdx < inputSize; ++bIdx) {
			if (aIdx == bIdx)
				continue;

			std::shared_ptr<Scanline3D> b = input[bIdx];

			std::shared_ptr<Scanline3D> aRoot = a->getRoot();
			std::shared_ptr<Scanline3D> bRoot = b->getRoot();

			// If they are the same or have the same root, skip:
			if (a == b || (aRoot == bRoot))
				continue;

			float* aStartPtr = &pcData[a->startIdx * 3];
			float* aEndPtr = &pcData[a->endIdx * 3];
			float* bStartPtr = &pcData[b->startIdx * 3];
			float* bEndPtr = &pcData[b->endIdx * 3];

			/* Faster implementation by doing it directly in this loop and avoiding function calls: */
			float aMeanX = (aStartPtr[0] + aEndPtr[0]) / 2;
			float aMeanY = (aStartPtr[1] + aEndPtr[1]) / 2;
			float aMeanZ = (aStartPtr[2] + aEndPtr[2]) / 2;

			float bMeanX = (bStartPtr[0] + bEndPtr[0]) / 2;
			float bMeanY = (bStartPtr[1] + bEndPtr[1]) / 2;
			float bMeanZ = (bStartPtr[2] + bEndPtr[2]) / 2;

			float dX = (aMeanX - bMeanX);
			float dY = (aMeanY - bMeanY);
			float dZ = (aMeanZ - bMeanZ);

			float distance = dX * dX + dY * dY + dZ * dZ;

			/* Previous Implementation, slower:
			FVector aStart(aStartPtr[0], aStartPtr[1], aStartPtr[2]);
			FVector aEnd(aEndPtr[0], aEndPtr[1], aEndPtr[2]);
			FVector bStart(bStartPtr[0], bStartPtr[1], bStartPtr[2]);
			FVector bEnd(bEndPtr[0], bEndPtr[1], bEndPtr[2]);

			FVector aMean = (aStart + aEnd) / 2;
			FVector bMean = (bStart + bEnd) / 2;

			float distance = (aMean - bMean).SizeSquared();
			*/

			// ALSO A PARAMETER OF THE ALGORITHM:
			if (distance < SCANLINE_MERGING_MAX_SQR_MIDPOINT_DISTANCE)
				a->Union(b);
		}
	}
}

/**
* Returns multiple point clouds with the mid points of the scanlines.
*
* Scanlines, which are connected and defines one region are merged in one point cloud.
*/

std::vector<boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>>> getSegmentedPointsFromScanlines(float* pcData, std::vector<std::shared_ptr<Scanline3D>> scanlines) {
	// First, assign all root scanlines an id:
	int segmentIdx = 0;
	std::vector<boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>>> result;

	for (std::shared_ptr<Scanline3D> scanline : scanlines) {
		if (scanline->parent == nullptr) {
			scanline->segmentIdx = segmentIdx++;
			result.push_back(boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>>(new pcl::PointCloud<pcl::PointXYZ>()));
		}
	}

	// Second, group all scanlines by there root id:
	for (std::shared_ptr<Scanline3D> scanline : scanlines) {
		float* startPtr = &pcData[scanline->startIdx * 3];
		float* endPtr = &pcData[scanline->endIdx * 3];

		FVector start(startPtr[0], startPtr[1], startPtr[2]);
		FVector end(endPtr[0], endPtr[1], endPtr[2]);
		FVector mid = (start + end) / 2;

		int idx = scanline->getRoot()->segmentIdx;

		result[idx]->push_back(pcl::PointXYZ(mid.X, mid.Y, mid.Z));
	}

	return result;
}


struct GridCandidate {
	int index;
	FVector x;
	FVector y;
	FVector z;
	FVector p;
};

/*
* Estimate grid candidates by the given points.
*/

TArray<GridCandidate> findGridCandidates(std::vector<boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>>> points) {
	TArray<GridCandidate> result;

	int currentIdx = 0;

	// Implemented like: http://codextechnicanum.blogspot.com/2015/04/find-minimum-oriented-bounding-box-of.html
	// Thank you for this example to make it a easier to get into the api of PCL!
	for (const boost::shared_ptr<const pcl::PointCloud<pcl::PointXYZ>> segmentCloud : points) {
		Eigen::Vector4f pcaCentroid;
		pcl::compute3DCentroid(*segmentCloud, pcaCentroid);
		Eigen::Matrix3f covariance;
		computeCovarianceMatrixNormalized(*segmentCloud, pcaCentroid, covariance);
		Eigen::SelfAdjointEigenSolver<Eigen::Matrix3f> eigen_solver(covariance, Eigen::ComputeEigenvectors);
		Eigen::Matrix3f eigenVectorsPCA = eigen_solver.eigenvectors();
		Eigen::Vector3f eigenValuesPCA = eigen_solver.eigenvalues();
		eigenVectorsPCA.col(2) = eigenVectorsPCA.col(0).cross(eigenVectorsPCA.col(1));

		Eigen::Matrix4f projectionTransform(Eigen::Matrix4f::Identity());
		projectionTransform.block<3, 3>(0, 0) = eigenVectorsPCA.transpose();
		projectionTransform.block<3, 1>(0, 3) = -1.f * (projectionTransform.block<3, 3>(0, 0) * pcaCentroid.head<3>());
		pcl::PointCloud<pcl::PointXYZ>::Ptr cloudPointsProjected(new pcl::PointCloud<pcl::PointXYZ>);
		pcl::transformPointCloud(*segmentCloud, *cloudPointsProjected, projectionTransform);

		pcl::PointXYZ minPoint, maxPoint;
		pcl::getMinMax3D(*cloudPointsProjected, minPoint, maxPoint);
		const Eigen::Vector3f meanDiagonal = 0.5f*(maxPoint.getVector3fMap() + minPoint.getVector3fMap());

		const Eigen::Quaternionf bboxQuaternion(eigenVectorsPCA);
		const Eigen::Matrix3f rot = bboxQuaternion.toRotationMatrix();

		FTransform rotation = FTransform(
			FVector(rot.coeff(0, 0), rot.coeff(1, 0), rot.coeff(2, 0)),
			FVector(rot.coeff(0, 1), rot.coeff(1, 1), rot.coeff(2, 1)),
			FVector(rot.coeff(0, 2), rot.coeff(1, 2), rot.coeff(2, 2)),
			FVector(0)
		);

		const Eigen::Vector3f bboxTransform = eigenVectorsPCA * meanDiagonal + pcaCentroid.head<3>();
		FVector bboxTransform4f(bboxTransform.x(), bboxTransform.y(), bboxTransform.z());

		// z and y have to be much greater than x:
		bool condition = eigenValuesPCA.z() / PCA_MIN_XY_TO_Z_RATIO > eigenValuesPCA.x() && eigenValuesPCA.y() / PCA_MIN_XY_TO_Z_RATIO > eigenValuesPCA.x();

		// there must be enough points in that segment:
		condition &= segmentCloud->size() > PCA_MIN_POINT_SIZE;

		// z and y should not differ that much:
		condition &= eigenValuesPCA.z() / PCA_MAX_X_TO_Y_RATION < eigenValuesPCA.y() && eigenValuesPCA.z() * PCA_MAX_X_TO_Y_RATION > eigenValuesPCA.y();

		FVector relativeMinPoint = rotation.TransformVector(FVector(minPoint.x, minPoint.y, minPoint.z));

		if (condition) {
			GridCandidate gridCandidate;
			gridCandidate.index = currentIdx;
			gridCandidate.x = (rotation.TransformVector(FVector(maxPoint.x, minPoint.y, minPoint.z)) - relativeMinPoint);
			gridCandidate.y = (rotation.TransformVector(FVector(minPoint.x, maxPoint.y, minPoint.z)) - relativeMinPoint);
			gridCandidate.z = (rotation.TransformVector(FVector(minPoint.x, minPoint.y, maxPoint.z)) - relativeMinPoint);
			gridCandidate.p = bboxTransform4f;

			if (gridCandidate.x.Normalize() && gridCandidate.y.Normalize() && gridCandidate.z.Normalize()) {
				result.Add(gridCandidate);
			}
		}

		++currentIdx;
	}

	return result;
}

/*
* Simply removes points which lies tight together to get rid of points which don't improve the result
* but results in bad performance in the next step.
*/

void reducePoints(boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>> points) {
	// Reduce a few points to reduce many point tests in the next step (performance):
	for (unsigned int i = 0; i < points->size(); ++i) {
		pcl::PointXYZ pI = (*points)[i];
		for (unsigned int h = i + 1; h < points->size(); ++h) {
			pcl::PointXYZ pH = (*points)[h];

			float dX = pI.x - pH.x;
			float dY = pI.y - pH.y;
			float dZ = pI.z - pH.z;

			// PARAM OF THE ALGORITHM: currently 2cm.
			if (dX * dX + dY * dY + dZ * dZ < REDUCE_POINTS_MAX_SQR_DISTANCE) {
				points->erase(points->begin() + h--);
			}
		}
	}
}

/**
* Filters all the points in the original point cloud which are near to the given points.
*
* PERFORAMNCE NOTE:
* Because all points of the point cloud (pcData) are tested against the given points (points), this 
* is currently the biggest bottleneck of the algorithm. TODO: Precalculate an approx. bounding box by the 
* PCA-results and use them instead all of the points (points) would propably more performant.
*/

void filterAllPointsInRadius(
	boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>> out_allInRadius,
	std::vector<int>& out_allInRadiusIndices,
	float* pcData, 
	unsigned int pcSize, 
	boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>> points
) {

	/* 
	 * NEW IMPLEMENTATION, HIGHLY INCREASES PERFORMANCE (BY FIRST CALCULATING A ROUGH BOUNDING SPHERE 
	 * TO AVOID TESTING SO MUCH POINTS AGAINST ALL POINTS OF THE POINT CLOUD) 
	 */

	// Calculate rough bounding sphere:
	pcl::PointXYZ min(100000, 100000, 100000);
	pcl::PointXYZ max(-100000, -100000, -100000);

	for (pcl::PointXYZ p : *points) {
		min.x = std::min(p.x, min.x);
		min.y = std::min(p.y, min.y);
		min.z = std::min(p.z, min.z);

		max.x = std::max(p.x, max.x);
		max.y = std::max(p.y, max.y);
		max.z = std::max(p.z, max.z);
	}

	float sphereMidX = (max.x + min.x) / 2;
	float sphereMidY = (max.y + min.y) / 2;
	float sphereMidZ = (max.z + min.z) / 2;
	float sphereSqrRad = 0;

	for (pcl::PointXYZ p : *points) {
		float dX = sphereMidX - p.x;
		float dY = sphereMidY - p.y;
		float dZ = sphereMidZ - p.z;

		float sqrDistance = dX * dX + dY * dY + dZ * dZ;
		sphereSqrRad = std::max(sphereSqrRad, sqrDistance);
	}

	// add max distance to sphereSqrRad:
	sphereSqrRad = std::sqrt(FILTER_PC_MAX_SQR_RADIUS) + std::sqrt(sphereSqrRad);
	sphereSqrRad = sphereSqrRad * sphereSqrRad;

	float maxDistancePTPSqr = FILTER_PC_MAX_SQR_RADIUS;

	// Now, for every point in the point cloud, first check against the rough boundin sphere, then against all other points:
	for (unsigned int i = 0; i < pcSize; ++i) {
		// First check against bounding sphere:
		float* pointPtr = &pcData[i * 3];

		float dX = sphereMidX - pointPtr[0];
		float dY = sphereMidY - pointPtr[1];
		float dZ = sphereMidZ - pointPtr[2];

		float distance = dX * dX + dY * dY + dZ * dZ;

		// If point is not in the bounding sphere (incl. additional margin), skip;
		if (distance > sphereSqrRad)
			continue;

		for (pcl::PointXYZ p : *points) {
			float d2X = p.x - pointPtr[0];
			float d2Y = p.y - pointPtr[1];
			float d2Z = p.z - pointPtr[2];

			float distance2 = d2X * d2X + d2Y * d2Y + d2Z * d2Z;

			if (distance2 <= maxDistancePTPSqr) {
				out_allInRadius->push_back(pcl::PointXYZ(pointPtr[0], pointPtr[1], pointPtr[2]));
				out_allInRadiusIndices.push_back(i);
				break;
			}
		}
	}

	/* Old implementation, uses no rough sphere bounding box. 
	auto sqrEuclDistance = [](pcl::PointXYZ a, pcl::PointXYZ b) {
		pcl::PointXYZ v(a.x - b.x, a.y - b.y, a.z - b.z);

		return v.x * v.x + v.y * v.y + v.z * v.z;
	};

	auto isInRadius = [sqrEuclDistance](boost::shared_ptr<const pcl::PointCloud<pcl::PointXYZ>> points, pcl::PointXYZ point) {
		for (pcl::PointXYZ p : *points) {
			if (sqrEuclDistance(p, point) < FILTER_PC_MAX_SQR_RADIUS) {
				return true;
			}
		}
		return false;
	};

	for (unsigned int i = 0; i < pcSize; ++i) {
		float* pointRaw = &pcData[i*3];
		pcl::PointXYZ point(pointRaw[0], pointRaw[1], pointRaw[2]);

		if (isInRadius(points, point)) {
			out_allInRadius->push_back(point);
			out_allInRadiusIndices.push_back(i);
		}
	}
	*/
}

/*
* Do the first ransac pass to filter wrong points (especially those which occurs by the flying pixel effect).
*/

void doFirstRansacPass(
	boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>> allInRadius,
	std::vector<int>& allInRadiusIndices,
	boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>> filteredInRadius_OUT,
	std::vector<int>& filteredInRadiusIndices_OUT,
	std::unordered_set<int>& filteredInRadiusIndicesSet_OUT
) {
	Eigen::VectorXf modelCoefficients;

	std::vector<int> inliers;
	pcl::SampleConsensusModelPlane<pcl::PointXYZ>::Ptr model_p(new pcl::SampleConsensusModelPlane<pcl::PointXYZ>(allInRadius));
	pcl::RandomSampleConsensus<pcl::PointXYZ> ransac(model_p);
	ransac.setDistanceThreshold(RANSAC_FIRST_PASS_THRESHOLD);
	ransac.computeModel();
	ransac.getInliers(inliers);
	ransac.getModelCoefficients(modelCoefficients);

	// copies all inliers of the model computed to another PointCloud
	pcl::copyPointCloud(*allInRadius, inliers, *filteredInRadius_OUT);

	for (int index : inliers) {
		filteredInRadiusIndices_OUT.push_back(allInRadiusIndices[index]);
		filteredInRadiusIndicesSet_OUT.insert(allInRadiusIndices[index]);
	}
}

/*
* Now, we search for precise holes in the filtered point cloud.
*/

std::vector<std::shared_ptr<Scanline3D>> findPreciseHoleScanlines(
	std::vector<int>& filteredInRadiusIndices, 
	std::unordered_set<int>& filteredInRadiusIndicesSet,
	unsigned int imgWidth
) {

	std::vector<std::pair<int, int>> firstAndLastIndexOfScanline;

	int lineStartIdx = -1;
	int lastIndex = -999999;

	// Finding pairs of start and end for each scanline:
	for (int index : filteredInRadiusIndices) {
		if (lastIndex / imgWidth != index / imgWidth) {
			if (lineStartIdx != -1) {
				firstAndLastIndexOfScanline.push_back(std::pair<int, int>(lineStartIdx, lastIndex));
			}

			lineStartIdx = index;
		}

		lastIndex = index;
	}


	std::vector<std::shared_ptr<Scanline3D>> holeScanlines;
	// Finding precise holes:
	{
		std::shared_ptr<Scanline3D> currentLine = nullptr;

		int lastI = -9999999;
		for (std::pair<int, int> pair : firstAndLastIndexOfScanline) {
			for (int i = pair.first; i <= pair.second; ++i) {

				if (!filteredInRadiusIndicesSet.count(i)) {
					if (i / imgWidth != lastI / imgWidth) {
						if (currentLine != nullptr) {
							holeScanlines.push_back(currentLine);
						}
						currentLine = std::shared_ptr<Scanline3D>(new Scanline3D(i));
					}
					else {
						// If next to each other, just connect:
						if (lastI == i - 1) {
							currentLine->endIdx = i;
						}
						else {
							holeScanlines.push_back(currentLine);
							currentLine = std::shared_ptr<Scanline3D>(new Scanline3D(i));
						}
					}
					lastI = i;
				}
			}
		}
	}

	return holeScanlines;
}

/*
* Merges neightboring scanlines (which are direct neighbours in the assumed depth image).
*/

void mergeDirectNeighBours(std::vector<std::shared_ptr<Scanline3D>> input, unsigned int imgWidth) {
	unsigned int i = 0;
	unsigned int j = 1;

	while (j < input.size()) {
		std::shared_ptr<Scanline3D> I = input[i];
		std::shared_ptr<Scanline3D> J = input[j];
		int iY = I->startIdx / imgWidth;
		int jY = J->startIdx / imgWidth;

		if (iY == jY) {
			++j;
		}
		else if (iY + 1 < jY) {
			++i;
		}
		else {
			int iXStart = I->startIdx % imgWidth;
			int iXEnd = I->endIdx % imgWidth;

			int jXStart = J->startIdx % imgWidth;
			int jXEnd = J->endIdx % imgWidth;

			// if scanlines of consecutive rows are overlapping (touching each other), try to merge:
			if (iXStart <= jXEnd && iXEnd >= jXStart) {
				input[i]->Union(input[j]);
			}

			// if both are in consecutive rows, go further with the scanline which the lower end:
			if (iXEnd < jXEnd)
				++i;
			else
				++j;
		}
	}
}

/**
* Second ransac pass with a lower threshold to estimate a more precise model.
*/

std::pair<FVector, FVector> doSecondRansacPass(boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>> allInRadius) {
	
	Eigen::VectorXf modelCoefficients;

	// Redo Ransac for a more precise model (:
	{
		
		pcl::SampleConsensusModelPlane<pcl::PointXYZ>::Ptr model_p(new pcl::SampleConsensusModelPlane<pcl::PointXYZ>(allInRadius->makeShared()));
		pcl::RandomSampleConsensus<pcl::PointXYZ> ransac(model_p);
		ransac.setDistanceThreshold(RANSAC_SECOND_PASS_THRESHOLD);
		ransac.computeModel();
		ransac.getModelCoefficients(modelCoefficients);
	
	}
	
	FVector planeN(modelCoefficients[0], modelCoefficients[1], modelCoefficients[2]);
	planeN.Normalize();

	FVector planeP(planeN.X * -modelCoefficients[3], planeN.Y * -modelCoefficients[3], planeN.Z * -modelCoefficients[3]);
	

	//return std::pair<FVector, FVector>(FVector(), FVector());
	return std::pair<FVector, FVector>(planeN, planeP);
}


FVector projectToPlane(FVector point, FVector planeP, FVector planeN) {
	FVector diff = point - planeP;
	float dist = FVector::DotProduct(diff, planeN);

	return point - planeN * dist;
}

/**
* Calculate the mid points of the holes.
*/

std::vector<FVector> calculateGridMidPoints(
	float* pcData, 
	unsigned int imgWidth,
	std::pair<FVector, FVector> plane,
	std::vector<std::shared_ptr<Scanline3D>>& holeScanlines
) {
	int segmentIdx = 0;
	for (std::shared_ptr<Scanline3D> scanline : holeScanlines) {
		if (scanline->parent == nullptr) {
			scanline->segmentIdx = segmentIdx++;
		}
	}

	std::vector<FVector> pointSum(segmentIdx);
	std::vector<int> pointSumCount(segmentIdx);
	std::vector<int> maxRowLength(segmentIdx);
	std::vector<int> rowCount(segmentIdx);
	std::vector<std::set<int>> visitedRows(segmentIdx);

	for (std::shared_ptr<Scanline3D> segment : holeScanlines) {
		int pIndex = segment->getRoot()->segmentIdx;

		int prevLastIdx = segment->startIdx - 1;
		int nextFirstIdx = segment->endIdx + 1;

		float* prevLastPtr = &pcData[prevLastIdx * 3];
		float* nextFirstPtr = &pcData[nextFirstIdx * 3];

		FVector prevLast(prevLastPtr[0], prevLastPtr[1], prevLastPtr[2]);
		FVector nextFirst(nextFirstPtr[0], nextFirstPtr[1], nextFirstPtr[2]);

		FVector first = projectToPlane(prevLast, plane.second, plane.first);
		FVector last = projectToPlane(nextFirst, plane.second, plane.first);
		FVector diff = last - first;

		int lengthMinus1 = segment->endIdx - segment->startIdx + 2;

		for (int i = 0; i <= lengthMinus1; ++i) {
			FVector p = first + diff * (i / float(lengthMinus1));

			pointSum[pIndex] = pointSum[pIndex] + p;
			++pointSumCount[pIndex];
		}

		maxRowLength[pIndex] = std::max(maxRowLength[pIndex], segment->endIdx - segment->startIdx + 1);
		if (visitedRows[pIndex].find(segment->startIdx / imgWidth) == visitedRows[pIndex].end()) {
			visitedRows[pIndex].insert(segment->startIdx / imgWidth);
			++rowCount[pIndex];
		}
	}

	// Calculate grid mid points:
	std::vector<FVector> gridMidPoints;
	for (unsigned int i = 0; i < pointSum.size(); ++i) {
		FVector midPoint = pointSum[i] / pointSumCount[i];

		if (rowCount[i] < GRID_MID_PIXELS_MAX_SIDE_LENGTH && maxRowLength[i] < GRID_MID_PIXELS_MAX_SIDE_LENGTH && pointSumCount[i] > GRID_MID_PIXEL_MIN_PIXEL_COUNT) {
			gridMidPoints.push_back(midPoint);
		}
	}

	return gridMidPoints;
}

FVector mean(std::vector<FVector> vectors) {
	FVector mean(0, 0, 0);
	int count = 0;
	for (FVector v : vectors) {
		mean = mean + v;
		++count;
	}
	return mean / count;
}

std::vector<FVector> filterByDistance(FVector reference, std::vector<FVector> vectors, float distance) {
	std::vector<FVector> result;
	for (FVector v : vectors) {
		if (FVector::Distance(v, reference) < distance) {
			result.push_back(v);
		}
	}

	return result;
}

/**
* This method specially returns clusters based on the angle to neighbouring hole mid points...
*/

std::vector<std::vector<std::tuple<float, int, int>>> clusterGridMidPointsByRotation(std::vector<FVector> gridMidPoints) {
	std::vector<std::tuple<float, int, int>> nearVectorDir;

	for (unsigned int i = 0; i < gridMidPoints.size(); ++i) {
		for (unsigned int h = 0; h < gridMidPoints.size(); ++h) {
			if (i == h)
				continue;

			FVector a = gridMidPoints[i];
			FVector b = gridMidPoints[h];

			FVector v = b - a;

			if (v.Size() < CLUSTER_MID_POINTS_MAX_DISTANCE && v.Size() > CLUSTER_MID_POINTS_MIN_DISTANCE) {
				float dir = atan2(v.Y, v.X);

				if (dir < 0)
					dir += M_PI;

				// map to 0 - PI:
				nearVectorDir.push_back(std::tuple<float, int, int>(dir, i, h));
			}
		}
	}

	// Sort by degree:
	std::sort(nearVectorDir.begin(), nearVectorDir.end(), [](std::tuple<float, int, int> a, std::tuple<float, int, int> b) {
		return std::get<0>(a) < std::get<0>(b);
	});

	std::vector<std::vector<std::tuple<float, int, int>>> clusters;
	clusters.push_back(std::vector<std::tuple<float, int, int>>());

	auto radDistanceForTuple = [](std::tuple<float, int, int> a, std::tuple<float, int, int> b) {
		float dist = abs(std::get<0>(a) - std::get<0>(b));
		return dist <= M_PI / 2 ? dist : M_PI - dist;
	};

	float MAX_RADIAL_MERGE_DISTANCE = CLUSTER_MID_POINTS_MAX_RADIAL_MERGING_DISTANCE;

	for (std::tuple<float, int, int> v : nearVectorDir) {
		std::vector<std::tuple<float, int, int>>& lastCluster = clusters.back();

		if (lastCluster.size() < 1) {
			lastCluster.push_back(v);
		}
		else if (radDistanceForTuple(lastCluster.back(), v) < MAX_RADIAL_MERGE_DISTANCE) {
			// if distance is low, add it to the cluster:
			lastCluster.push_back(v);
		}
		else {
			// if distance is higher, create a new cluster:
			clusters.push_back(std::vector<std::tuple<float, int, int>>());
			clusters.back().push_back(v);
		}
	}

	// Merge last and first vector:
	if (clusters.size() > 1 && radDistanceForTuple(clusters.back().back(), clusters.front().front()) < MAX_RADIAL_MERGE_DISTANCE) {
		for (std::tuple<float, int, int> v : clusters.back()) {
			clusters.front().push_back(v);
		}
		clusters.pop_back();
	}

	int i = 0;
	for (std::vector<std::tuple<float, int, int>> cluster : clusters) {
		for (std::tuple<float, int, int> line : cluster) {

		}
		++i;
	}

	std::sort(clusters.begin(), clusters.end(), [](std::vector<std::tuple<float, int, int>> a, std::vector<std::tuple<float, int, int>> b) {
		return a.size() > b.size();
	});

	return clusters;
}


/**
* This method removes wrong points by analysing the two biggest clusters of the result of the clusterGridMidPointsByRotation-method.
*/

std::vector<FVector> filterByTwoMajorClusters(std::vector<std::vector<std::tuple<float, int, int>>> clusters, std::vector<FVector> vectors, int minClusterCount) {
	int* gridMidPointGoodIndicator = new int[vectors.size()]{ 0 };
	int* gridMidPointBadIndicator = new int[vectors.size()]{ 0 };

	for (unsigned int clusterID = 0; clusterID < 2 && clusters.size(); ++clusterID) {
		std::vector<std::tuple<float, int, int>>& cluster = clusters[clusterID];

		for (std::tuple<float, int, int>& line : cluster) {
			// Increase the counting array for all points those outgoing vectors are in:
			++gridMidPointGoodIndicator[std::get<1>(line)];
		}
	}

	for (unsigned int clusterID = 2; clusterID < clusters.size(); ++clusterID) {
		std::vector<std::tuple<float, int, int>>& cluster = clusters[clusterID];

		for (std::tuple<float, int, int>& line : cluster) {
			// Increase the counting array for all points those outgoing vectors are in:
			--gridMidPointGoodIndicator[std::get<1>(line)];
		}
	}

	std::vector<FVector> result;
	for (unsigned int i = 0; i < vectors.size(); ++i) {
		if (gridMidPointGoodIndicator[i] >= minClusterCount && gridMidPointGoodIndicator[i] > gridMidPointBadIndicator[i]) {
			result.push_back(vectors[i]);
		}
	}

	delete[] gridMidPointGoodIndicator;
	delete[] gridMidPointBadIndicator;
	return result;
}


int findNearestPointTo(FVector gridMidWeighted, std::vector<FVector> gridMidPoints) {
	if (gridMidPoints.size() == 0)
		return -1;

	int result = 0;
	float distance = FVector::Dist(gridMidPoints[0], gridMidWeighted);

	for (unsigned int i = 1; i < gridMidPoints.size(); ++i) {
		float current_dist = FVector::Dist(gridMidPoints[i], gridMidWeighted);

		if (current_dist < distance) {
			distance = current_dist;
			result = i;
		}
	}

	return result;
}

std::vector<FVector> filterByAxisDistance(FVector reference, std::vector<FVector> vectors, FVector axis1, FVector axis2, float maxDist) {
	std::vector<FVector> result;

	for (FVector v : vectors) {
		FVector diff = v - reference;

		if (FVector::DotProduct(diff,axis1) <= maxDist && FVector::DotProduct(diff, axis2) <= maxDist) {
			result.push_back(v);
		}
	}

	return result;
}


std::vector<int> filterPointsAround(float* pcData, unsigned int pcSize, FVector gridCenter, FVector axis1, FVector axis2, FVector normal, float minDist, float maxDist, float maxNormalDist) {
	std::vector<int> result;

	for (unsigned int i = 0; i < pcSize; ++i) {
		float* pc = &pcData[i*3];

		float pX = pc[0] - gridCenter.X;
		float pY = pc[1] - gridCenter.Y;
		float pZ = pc[2] - gridCenter.Z;

		if (isnan(pX) || isnan(pY) || isnan(pZ))
			continue;

		/* Faster implementation (because this is executed for all points of the original point cloud): */
		float axis1Distance = abs(pX * axis1.X + pY * axis1.Y + pZ * axis1.Z);
		float axis2Distance = abs(pX * axis2.X + pY * axis2.Y + pZ * axis2.Z);
		float normalDistance = abs(pX * normal.X + pY * normal.Y + pZ * normal.Z);

		/* Old implementation, slower:
		FVector v(pc[0], pc[1], pc[2]);

		FVector p = v - gridCenter;

		if (p.ContainsNaN())
			continue;

		float axis1Distance = abs(FVector::DotProduct(p, axis1));
		float axis2Distance = abs(FVector::DotProduct(p, axis2));
		float normalDistance = abs(FVector::DotProduct(p, normal));
		*/

		if (normalDistance > maxNormalDist)
			continue;

		if (axis1Distance < minDist && axis2Distance < minDist)
			continue;

		if (axis1Distance > maxDist || axis2Distance > maxDist)
			continue;

		result.push_back(i);
	}

	return result;
}

/**
* Detects a grid in the given point cloud and returns the result.
*/

GridDetectionResult GridDetector::detectGrid(PointCloudImage pointCloud) {
	auto startTimeMeasure = std::chrono::system_clock::now();

	// Filter connected scanlines:
	std::vector<std::shared_ptr<Scanline3D>> scanlines = filterConnectedScanlines(pointCloud.data, pointCloud.sourceImageWidth, pointCloud.sourceImageHeight);
	
	// Filter holes in straight lines:
	scanlines = filterHolesInStraigtLines(pointCloud.data, pointCloud.sourceImageWidth, scanlines);

	// Merge scanlines which are near to each other (it is assumed that scanlines which are merged are part of the same grid):
	mergeNeighbours(pointCloud.data, scanlines);

	// Create clustered point clouds from the given scanlines (by their mid points):
	std::vector<boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>>> clusteredPoints = getSegmentedPointsFromScanlines(pointCloud.data, scanlines);

	// Find grid candidates based on the clustered point clouds:
	TArray<GridCandidate> gridCandidates = findGridCandidates(clusteredPoints);

	// For each candidate:
	for (GridCandidate gridCandidate : gridCandidates) {
		// Reduce points which are very near to each other to make the "filterAllPointsInRadius"-method faster.
		reducePoints(clusteredPoints[gridCandidate.index]);

		// Initialize data structures for next function call:
		boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>> allInRadius = boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>>(new pcl::PointCloud<pcl::PointXYZ>());
		std::vector<int> allInRadiusIndices;

		// Filters all points in a given radius and inserts them into allInRadius and their indices in the pointCloud.data into allInRadiusIndices:
		filterAllPointsInRadius(allInRadius, 
			allInRadiusIndices, pointCloud.data, 
			pointCloud.sourceImageWidth * pointCloud.sourceImageHeight, 
			clusteredPoints[gridCandidate.index]
		);

		// If there are too less points in radius, this can't be a grid:
		if (allInRadius->size() < FILTER_PC_MIN_SIZE_OF_GRID_CANDIDATE)
			continue;

		// Initialize data structures for next function call:
		boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>> filteredInRadius = boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>>(new pcl::PointCloud<pcl::PointXYZ>());
		std::vector<int> filteredInRadiusIndices;
		std::unordered_set<int> filteredInRadiusIndicesSet;

		// Do first ransac pass to estimate plane of the grid:
		doFirstRansacPass(allInRadius, allInRadiusIndices, filteredInRadius, filteredInRadiusIndices, filteredInRadiusIndicesSet);

		// Get scanlines of the precise holes:
		std::vector<std::shared_ptr<Scanline3D>> scanlinesOfHoles = 
			findPreciseHoleScanlines(filteredInRadiusIndices, filteredInRadiusIndicesSet, pointCloud.sourceImageWidth);
		
		// Merge this holes:
		mergeDirectNeighBours(scanlinesOfHoles, pointCloud.sourceImageWidth);

		// Estimate plane normal and plane position by a second ransac pass (with a smaller inlier threshold for more precision):
		std::pair<FVector, FVector> plane = doSecondRansacPass(allInRadius);
		FVector planeN = plane.first;
		FVector planeP = plane.second;

		// Make plane always point in the same z direction:
		if (planeN.Z > 0)
			planeN = planeN * -1;

		std::vector<FVector> gridMidPoints =
			calculateGridMidPoints(pointCloud.data, pointCloud.sourceImageWidth, std::pair<FVector, FVector>(planeN, planeP), scanlinesOfHoles);

		// Generate rotation-clusters of near-by vectors:
		std::vector<std::vector<std::tuple<float, int, int>>> clusters = clusterGridMidPointsByRotation(gridMidPoints);

		// if no clear clusters where found, don't recognize the grid:
		if (clusters.size() < 2 || clusters[0].size() < CLUSTER_MID_POINTS_MIN_RADIAL_CLUSTER_SIZE || clusters[1].size() < CLUSTER_MID_POINTS_MIN_RADIAL_CLUSTER_SIZE) {
			continue;
		}

		// Get first axis by our radial cluster heuristic:
		std::tuple<float, int, int> cluster0Mean = clusters[0][clusters[0].size() / 2];
		FVector axis1 = (gridMidPoints[std::get<2>(cluster0Mean)] - gridMidPoints[std::get<1>(cluster0Mean)]);
		axis1.Normalize();

		// get second axis by our radial cluster heuristic:
		std::tuple<float, int, int> cluster1Mean = clusters[1][clusters[1].size() / 2];
		FVector axis2 = (gridMidPoints[std::get<2>(cluster1Mean)] - gridMidPoints[std::get<1>(cluster1Mean)]);
		axis2.Normalize();

		// Filter all points which are not connected to another point over minimal 1 good angle:
		gridMidPoints = filterByTwoMajorClusters(clusters, gridMidPoints, 1);

		unsigned int MIN_GRID_POINTS = 16; // Should also be a parameter.
		if (gridMidPoints.size() < MIN_GRID_POINTS)
			continue;

		// Calculate weighted:
		FVector gridMidWeighted = mean(gridMidPoints);

		// Only use grid points in specific radius for more precise mid prediction:
		std::vector<FVector> midPointsInRadius = filterByDistance(gridMidWeighted, gridMidPoints, MID_POINT_PREDICTION_MAXIMAL_RADIUS);

		// Calculate again:
		gridMidWeighted = mean(midPointsInRadius);

		// Find nearest point to the weighted grid mid and use it at estimated grid mid (to remove imprecision if some hole mid points
		// are not recognized):
		int nearestPointTo = findNearestPointTo(gridMidWeighted, midPointsInRadius);
		if (nearestPointTo < 0)
			continue;

		FVector probableGridMid = midPointsInRadius[nearestPointTo];

		gridMidPoints = filterByAxisDistance(probableGridMid, gridMidPoints, axis1, axis2, CLUSTER_MID_POINTS_MAXIMAL_AXIS_DISTANCE_TO_MID);
		
		if (gridMidPoints.size() == 0)
			continue;


		float axis1Length = 0;
		float axis2Length = 0;

		// Estimate length along the axes by projecting the points to them:
		{
			float minAxis1 = 1000000;
			float maxAxis1 = -1000000;

			float minAxis2 = 1000000;
			float maxAxis2 = -1000000;

			for (FVector mid : gridMidPoints) {
				float axis1Value = FVector::DotProduct(mid, axis1);
				float axis2Value = FVector::DotProduct(mid, axis2);

				minAxis1 = std::min(axis1Value, minAxis1);
				maxAxis1 = std::max(axis1Value, maxAxis1);

				minAxis2 = std::min(axis2Value, minAxis2);
				maxAxis2 = std::max(axis2Value, maxAxis2);
			}

			axis1Length = maxAxis1 - minAxis1;
			axis2Length = maxAxis2 - minAxis2;
		}

		float MAX_EDGE_LENGTH_OF_OBB = 0.36;
		float MIN_EDGE_LENGTH_OF_OBB = 0.28f;
		float MAX_GRID_MID_TO_WEIGHTED_DISTNACE = 0.03f;

		// Only continue, if the minimal bounding box has the size of a grid:
		if (FVector::Distance(gridMidWeighted, probableGridMid) < MAX_GRID_MID_TO_WEIGHTED_DISTNACE
			&& axis1Length > MIN_EDGE_LENGTH_OF_OBB && axis2Length > MIN_EDGE_LENGTH_OF_OBB
			&& axis1Length < MAX_EDGE_LENGTH_OF_OBB && axis2Length < MAX_EDGE_LENGTH_OF_OBB)
		{
			// Filter points AROUND the grid for orientation estimation (these points are normally hands and arms which hold the grid):
			std::vector<int> pointsAroundGridIndices = filterPointsAround(pointCloud.data, pointCloud.sourceImageWidth * pointCloud.sourceImageHeight, probableGridMid, axis1, axis2, planeN, 0.25f, 0.4f, 0.10f);

			FVector midPointOfPointsAroundGrid(0);

			for (int i : pointsAroundGridIndices) {
				float* pc = &pointCloud.data[i * 3];
				midPointOfPointsAroundGrid = midPointOfPointsAroundGrid + FVector(pc[0], pc[1], pc[2]);
			}

			midPointOfPointsAroundGrid = midPointOfPointsAroundGrid / pointsAroundGridIndices.size();

			GridDetectionResult result;
			result.gridFound = true;
			result.gridCenter = probableGridMid;

			FVector normAxis1 = axis1; normAxis1.Normalize();
			FVector normAxis2 = FVector::CrossProduct(normAxis1, planeN); normAxis2.Normalize();

			result.normal = planeN;
			result.vectorToHandsMidpoint = midPointOfPointsAroundGrid - probableGridMid;

			FVector vectorToHandsMidpointNormalized = result.vectorToHandsMidpoint; vectorToHandsMidpointNormalized.Normalize();

			float dPTB[4];
			dPTB[0] = FVector::DotProduct(vectorToHandsMidpointNormalized, normAxis1);
			dPTB[1] = FVector::DotProduct(vectorToHandsMidpointNormalized, normAxis2);
			dPTB[2] = FVector::DotProduct(vectorToHandsMidpointNormalized, -normAxis1);
			dPTB[3] = FVector::DotProduct(vectorToHandsMidpointNormalized, -normAxis2);

			// Orient the axis so that the x-axis directs to the midpoint of the hands:
			if (dPTB[0] > dPTB[1] && dPTB[0] > dPTB[2] && dPTB[0] > dPTB[3]) {
				result.axis1 = normAxis1;
				result.axis2 = normAxis2;
			}
			else if (dPTB[1] > dPTB[2] && dPTB[1] > dPTB[3]) {
				result.axis1 = normAxis2;
				result.axis2 = -normAxis1;
			}
			else if (dPTB[2] > dPTB[3]) {
				result.axis1 = -normAxis1;
				result.axis2 = -normAxis2;
			}
			else {
				result.axis1 = -normAxis2;
				result.axis2 = normAxis1;
			}

			for (FVector holePoints : gridMidPoints) {
				result.centersOfFoundHoles.Add(holePoints);
			}


			float totalRuntime = float((std::chrono::duration<double>(std::chrono::system_clock::now() - startTimeMeasure)).count() * 1000.f);
			UE_LOG(LogTemp, Warning, TEXT("Time Needed for Grid detection: %f"), totalRuntime)
			return result;
		}
	}


	float totalRuntime = float((std::chrono::duration<double>(std::chrono::system_clock::now() - startTimeMeasure)).count() * 1000.f);
	UE_LOG(LogTemp, Warning, TEXT("Time Needed for Grid detection: %f"), totalRuntime)
	return GridDetectionResult();
}