//  2021, CGVR (https://cgvr.informatik.uni-bremen.de/), Andre Mhlenbrock (muehlenb@uni-bremen.de)


#include "MultiplePointCloudsRegistrator.h"

THIRD_PARTY_INCLUDES_START
#include <pcl/io/pcd_io.h>
#include <pcl/point_types.h>
#include <pcl/registration/transformation_estimation.h>
#include <pcl/registration/transformation_estimation_svd.h>
#include <pcl/registration/correspondence_rejection_sample_consensus.h>
THIRD_PARTY_INCLUDES_END


MultiplePointCloudsRegistrator::MultiplePointCloudsRegistrator()
	: pointCloudCount(0) {}

MultiplePointCloudsRegistrator::MultiplePointCloudsRegistrator(unsigned int pPointCloudCount) 
	: pointCloudCount(pPointCloudCount) 
{
	// Initialize the array with the number of point clouds:
	for (unsigned int i = 0; i < pPointCloudCount; ++i) {
		correspondences.Add(TArray<FVector>());
		meanErrors.Add(-1);
	}
}

MultiplePointCloudsRegistrator::~MultiplePointCloudsRegistrator()
{
}

void MultiplePointCloudsRegistrator::addCorrespondencePoints(TArray<TArray<FVector>> points) {
	if (points.Num() == 0) {
		UE_LOG(LogTemp, Warning, TEXT("No points given (addCorrespondencePoints)!"));
		return;
	}

	if (points.Num() != pointCloudCount) {
		UE_LOG(LogTemp, Warning, TEXT("The outer array must match the size of the specified point clouds (addCorrespondencePoints)!"));
		return;
	}

	int sizeOfFirstArray = points[0].Num();
	for (int i = 1; i < points.Num(); ++i) {
		if (points[i].Num() != sizeOfFirstArray) {
			UE_LOG(LogTemp, Warning, TEXT("Arrays of different size are given (addCorrespondencePoints)!"));
			return;
		}
	}
	
	// Add all points:
	for (int i = 0; i < points.Num(); ++i) {
		for (FVector v : points[i]) {
			correspondences[i].Add(v);
		}
	}
}

FTransform pclMatrixToFTransform(pcl::registration::TransformationEstimationSVD<pcl::PointXYZ, pcl::PointXYZ>::Matrix4 mat) {
	FVector x = FVector(mat(0, 0), mat(1, 0), mat(2, 0));
	FVector y = FVector(mat(0, 1), mat(1, 1), mat(2, 1));
	FVector z = FVector(mat(0, 2), mat(1, 2), mat(2, 2));
	FVector trl = FVector(mat(0, 3), mat(1, 3), mat(2, 3));

	return FTransform(x, y, z, trl);
}

TArray<FTransform> MultiplePointCloudsRegistrator::registerInto(int pointCloudID, float correspondenceRejectionThreshold) {
	if (pointCloudCount < 2) {
		UE_LOG(LogTemp, Warning, TEXT("The number of specified point clouds has to be 2 at minimum! Was %i"), pointCloudCount);
		return TArray<FTransform>();
	}

	if (pointCloudID < 0 || pointCloudID >= correspondences.Num()) {
		UE_LOG(LogTemp, Warning, TEXT("The given point cloud id doesn't exists. Given point cloud id: %i. Size of point clouds: %i."), pointCloudID, pointCloudCount);
		return TArray<FTransform>();
	}

	if (correspondences[pointCloudID].Num() < 4) {
		UE_LOG(LogTemp, Warning, TEXT("There have to be at least 4 reference points for the specified point cloud id! There were just %i"), correspondences[pointCloudID].Num());
		return TArray<FTransform>();
	}

	TArray<FTransform> result;
	TArray<TArray<int>> tmpUsedCorrespondenceIndices;

	for (int i = 0; i < pointCloudCount; ++i) {
		tmpUsedCorrespondenceIndices.Add(TArray<int>());
	}

	TArray<int> numOfUsedCP;
	TArray<int> numOfTotalCP;

	for (int i = 0; i < pointCloudCount; ++i) {
		if (i == pointCloudID) {
			result.Add(FTransform());
			continue;
		}

		// Create point clouds (expected struct by pcl for correspondence points):
		boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>> sourcePC(new pcl::PointCloud<pcl::PointXYZ>());
		boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>> targetPC(new pcl::PointCloud<pcl::PointXYZ>());

		boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>> sourcePCFiltered(new pcl::PointCloud<pcl::PointXYZ>()); // Result of Correspondence Rejection
		boost::shared_ptr<pcl::PointCloud<pcl::PointXYZ>> targetPCFiltered(new pcl::PointCloud<pcl::PointXYZ>()); // Result of Correspondence Rejection

		// Copy all points into sourcePC and targetPC, where both vectors contains NO NAN value:
		for (int h = 0; h < correspondences[pointCloudID].Num(); ++h) {
			FVector srcPoint = correspondences[i][h];
			FVector tgtPoint = correspondences[pointCloudID][h];

			if (!srcPoint.ContainsNaN() && !tgtPoint.ContainsNaN()) {
				sourcePC->push_back(pcl::PointXYZ(srcPoint.X * 100, srcPoint.Y * 100, srcPoint.Z * 100));
				targetPC->push_back(pcl::PointXYZ(tgtPoint.X * 100, tgtPoint.Y * 100, tgtPoint.Z * 100));
				//UE_LOG(LogTemp, Warning, TEXT("Added P1: %f, %f, %f AND P2: %f, %f, %f"), srcPoint.X, srcPoint.Y, srcPoint.Z, tgtPoint.X, tgtPoint.Y, tgtPoint.Z)
			}
		}

		// Correspondence rejection:
		pcl::Correspondences filteredCorresp;
		{
			pcl::Correspondences corresp;

			for (int h = 0; h < sourcePC->size(); ++h) {
				corresp.push_back(pcl::Correspondence(h, h, std::numeric_limits<float>::max()));
			}

			boost::shared_ptr<pcl::Correspondences> corPtr(new pcl::Correspondences(corresp));

			pcl::registration::CorrespondenceRejectorSampleConsensus<pcl::PointXYZ> crsac;
			crsac.setInputSource(sourcePC);
			crsac.setInputTarget(targetPC);
			crsac.setInputCorrespondences(corPtr);
			crsac.setInlierThreshold(correspondenceRejectionThreshold * 100);
			crsac.setMaximumIterations(1000);
			crsac.getRemainingCorrespondences(corresp, filteredCorresp);

			for (pcl::Correspondence correspondence : filteredCorresp) {
				sourcePCFiltered->push_back((*sourcePC)[correspondence.index_query]);
				targetPCFiltered->push_back((*targetPC)[correspondence.index_match]);

				tmpUsedCorrespondenceIndices[i].Add(correspondence.index_query);
			}

			numOfUsedCP.Add(filteredCorresp.size());
			numOfTotalCP.Add(corresp.size());
		}

		// If there are not enough points, add simply identity:
		if (sourcePCFiltered->size() < 4 || targetPCFiltered->size() < 4) {
			result.Add(FTransform());
			continue;
		}

		pcl::registration::TransformationEstimationSVD<pcl::PointXYZ, pcl::PointXYZ> estimator;
		pcl::registration::TransformationEstimationSVD<pcl::PointXYZ, pcl::PointXYZ>::Matrix4 mat;
		estimator.estimateRigidTransformation(*sourcePCFiltered, *targetPCFiltered, mat);

		FTransform transform = pclMatrixToFTransform(mat);

		// Calculate error:
		{
			float sumError = 0;
			for (pcl::Correspondence correspondence : filteredCorresp) {
				pcl::PointXYZ sP = (*sourcePC)[correspondence.index_query];
				pcl::PointXYZ tP = (*targetPC)[correspondence.index_match];


				FVector sV(sP.x, sP.y, sP.z);
				FVector tV(tP.x, tP.y, tP.z);

				FVector transformedKinectPoint = transform.TransformPosition(sV);
				sumError += FVector::Distance(transformedKinectPoint, tV);
			}
			meanErrors[i] = sumError / filteredCorresp.size();
		}



		result.Add(transform);
	}

	lastUsedCorrespondencePointSize = numOfUsedCP;
	lastTotalCorrespondencePointSize = numOfTotalCP;

	return result;
}

float MultiplePointCloudsRegistrator::getMeanError(int pointCloudIndex) {
	return meanErrors[pointCloudIndex];
}