//  2021, CGVR (https://cgvr.informatik.uni-bremen.de/), Andre Mhlenbrock (muehlenb@uni-bremen.de)

#pragma once

#include "CoreMinimal.h"

#include "RegistrationStructs.h"

/**
* This class matches correspondence points between multiple GridDetectionResults.
*/

class VRKINECTREGISTRATION_API CorrespondencePointMatcher
{
public:

	/**
	* Matches all the hole centers of the grid detection result and returns for each GridDetectionResult
	* an array with the matched hole centers (that means that a specific hole center has the same index
	* in each array, considering the orientation). Every array of hole centers will contain 25 entries
	* in case of a 5x5 grid. Hole centers which were not detected will result in a FVector filled with
	* NANs.
	*
	* USAGE HINT:
	* Using the points of this method for registration should result in a faster convergence to a good
	* registration, but should only be used when the grid will be visible from the same side by all cameras
	* (e.g. when the depth image cameras are at the ceiling and look down to the grid).
	*
	* If some cameras capturing the front side and some the back side of the grid, PLEASE USE THE
	* onlyUseGridCenters-function.
	*/

	static TArray<TArray<FVector>> matchAllGridHoles_AssumingSameSides(TArray<GridDetectionResult> results);

	/**
	* Matches all the hole centers of the grid detection result and returns for each GridDetectionResult
	* an array with the matched hole centers (that means that a specific hole center has the same index
	* in each array, considering the orientation). Every array of hole centers will contain 25 entries
	* in case of a 5x5 grid. Hole centers which were not detected will result in a FVector filled with
	* NANs.
	*
	* To resolve the ambiguity that the sign of the Z-axis of the grid is not known, a rough registration
	* is assumed in advance. With the help of this rough registration it is checked whether the direction 
	* of the Z-axis is in the same or opposite direction compared to a reference camera (camera 0 is simply
	* used as reference camera).
	*/

	static TArray<TArray<FVector>> matchAllGridHoles_ByRoughRegistration(TArray<FTransform> roughRegistration, TArray<GridDetectionResult> results, int referenceID);

	/**
	* Matching Grid Center and a shifted grid center along the x-axis for a faster Registration than only
	* grid center.
	*/

	static TArray<TArray<FVector>> matchGridCenterAndXAxis(TArray<GridDetectionResult> results);

	/**
	 * Returns the grid centers of the GridDetectionResults.
	 */

	static TArray<TArray<FVector>> matchOnlyGridCenters(TArray<GridDetectionResult> results);

	/*
	* Finds the motion controller positions in a given point cloud by finding the grid and use
	* assumptions about where the motion controllers are relative to the grid.
	*/

	static MotionControllerResult findMotionControllerRelativeToGrid(GridDetectionResult, MotionControllerAssumptions controllerAssumptions);

	/*
	* Returns an array, first for the left and right motion controller in world space and second for the left and right motion controller
	* in space of the point cloud of the grid detection result.
	*/

	static TArray<TArray<FVector>> matchControllerAndGridDetectionResult(FVector lefControllerWS, FVector rightControllerWS, MotionControllerResult mcr);

};