// Fill out your copyright notice in the Description page of Project Settings.

#include "AzureKinect.h"

#include <chrono>

// Sets default values
AAzureKinect::AAzureKinect()
{
 	// Set this actor to call Tick() every frame.  You can turn this off to improve performance if you don't need it.
	PrimaryActorTick.bCanEverTick = true;

	RootComponent = CreateDefaultSubobject<USceneComponent>("RootComponent");
	pcRendererComponent = CreateDefaultSubobject<UPointCloudRenderer>("PCRendererComponent");
	pcRendererComponent->SetupAttachment(RootComponent);

	gridMeshComponent = CreateDefaultSubobject<UStaticMeshComponent>("GridMeshComponent");
	gridMeshComponent->SetupAttachment(RootComponent);

	estimatedLeftController = CreateDefaultSubobject<UStaticMeshComponent>("EstimatedLeftDebugComponent");
	estimatedLeftController->SetupAttachment(RootComponent);

	estimatedRightController = CreateDefaultSubobject<UStaticMeshComponent>("EstimatedRightDebugComponent");
	estimatedRightController->SetupAttachment(RootComponent);

	static ConstructorHelpers::FObjectFinder<UStaticMesh> GridMeshFinder(TEXT("StaticMesh'/Game/Geometry/Meshes/Grid.Grid'"));
	if (GridMeshFinder.Succeeded()) {
		gridMeshComponent->SetStaticMesh(GridMeshFinder.Object);
	}

	static ConstructorHelpers::FObjectFinder<UStaticMesh> SphereMeshFinder(TEXT("StaticMesh'/Engine/BasicShapes/Sphere.Sphere'"));
	if (GridMeshFinder.Succeeded()) {
		estimatedLeftController->SetStaticMesh(SphereMeshFinder.Object);
		estimatedLeftController->SetRelativeScale3D(FVector(0.1f));
		estimatedRightController->SetStaticMesh(SphereMeshFinder.Object);
		estimatedRightController->SetRelativeScale3D(FVector(0.1f));
	}
}

AAzureKinect::~AAzureKinect() {
}

bool AAzureKinect::IsInitialized() {
	return bIsInitialized;
}

void AAzureKinect::CreateXYLookupTable() {
	xyLookupTable = k4a::image::create(K4A_IMAGE_FORMAT_CUSTOM, width, height, width * (int)sizeof(k4a_float2_t));

	k4a_float2_t *table_data = (k4a_float2_t *)(void *)xyLookupTable.get_buffer();

	k4a_float2_t p;
	k4a_float3_t ray;

	for (unsigned int y = 0, idx = 0; y < height; y++)
	{
		p.xy.y = (float)y;

		for (unsigned int x = 0; x < width; x++, idx++)
		{
			p.xy.x = (float)x;

			if (sensorCalibration.convert_2d_to_3d(p, 1.f, K4A_CALIBRATION_TYPE_DEPTH, K4A_CALIBRATION_TYPE_DEPTH, &ray)) {
				table_data[idx].xy.x = ray.xyz.x;
				table_data[idx].xy.y = ray.xyz.y;
			}
			else {
				table_data[idx].xy.x = nanf("");
				table_data[idx].xy.y = nanf("");
			}
		}
	}
	UE_LOG(LogTemp, Warning, TEXT("Lookup Table successfully initialized!"));
}

bool AAzureKinect::Init(int deviceID)
{
	iDeviceID = deviceID;

	try
	{
		// Open kinect device and set calibration:
		nativeKinectDevice = k4a::device::open(deviceID);

		k4a_device_configuration_t deviceConfig = K4A_DEVICE_CONFIG_INIT_DISABLE_ALL;
		deviceConfig.depth_mode = K4A_DEPTH_MODE_NFOV_UNBINNED;
		deviceConfig.color_resolution = K4A_COLOR_RESOLUTION_720P;
		deviceConfig.color_format = K4A_IMAGE_FORMAT_COLOR_BGRA32;
		deviceConfig.synchronized_images_only = true;
		deviceConfig.camera_fps = K4A_FRAMES_PER_SECOND_30;

		nativeKinectDevice.start_cameras(&deviceConfig);

		sensorCalibration = nativeKinectDevice.get_calibration(deviceConfig.depth_mode, deviceConfig.color_resolution);
		deviceTransform = k4a::transformation(sensorCalibration);

		// Store width and height:
		width = sensorCalibration.depth_camera_calibration.resolution_width;
		height = sensorCalibration.depth_camera_calibration.resolution_height;

		// Initialize temporary arrays for point cloud and color data:
		tmpPointCloud = new float[width * height * 4];
		tmpColorImage = new uint8_t[width * height * 4];

		// Initialize Point cloud renderer:
		if (!pcRendererComponent->IsInitialized()) {
			// The point cloud renderer directly renders the temporary point cloud data (without mutex or lock):
			pcRendererComponent->Init(tmpPointCloud, tmpColorImage, width, height, iDeviceID);
		}

		// Create Lookup Table if not initialized yet:
		CreateXYLookupTable();

		bIsInitialized = true;


		UE_LOG(LogTemp, Warning, TEXT("Kinect %d initialized!"), deviceID);

		thread = new FAzureKinectThread((AAzureKinect*)this);

		return true;
	}
	catch (k4a::error initError)
	{
		FString initErrorStr(initError.what());
		UE_LOG(LogTemp, Error, TEXT("Error while initialize kinect %d: %s"), deviceID, *initErrorStr);
	}

	return false;
}

bool AAzureKinect::Shutdown() {
	if (thread)
		thread->Terminate();

	if (nativeKinectDevice)
	{
		nativeKinectDevice.stop_cameras();
		nativeKinectDevice.close();
		nativeKinectDevice = nullptr;
		UE_LOG(LogTemp, Warning, TEXT("Kinect %d closed!"), iDeviceID);
	}

	return true;
}

// Called when the game starts or when spawned
void AAzureKinect::BeginPlay()
{
	Super::BeginPlay();
}

bool AAzureKinect::ReadImage() {
	//auto startTime = std::chrono::high_resolution_clock::now();

	k4a::capture capture;

	bool gotCapture = false;
	try {
		gotCapture = nativeKinectDevice.get_capture(&capture);
	} catch (...) {
		UE_LOG(LogTemp, Error, TEXT("Error while reading the capture (try-catch)."));
		return false;
	}

	if (!gotCapture) {
		UE_LOG(LogTemp, Error, TEXT("Error while reading the capture (returned false)."));
		return false;
	}

	// Get color and depth image of kinect:
	k4a::image depthImage = capture.get_depth_image();
	k4a::image colorImage = capture.get_color_image();

	// registrate color image into depth image:
	k4a::image transformedColorImage = deviceTransform.color_image_to_depth_camera(depthImage, colorImage);

	if (depthImage == NULL) {
		UE_LOG(LogTemp, Error, TEXT("Depth image is null!"));
		return false;
	}
	
	if(transformedColorImage == NULL) {
		UE_LOG(LogTemp, Error, TEXT("Transformed color image is null!"));
		return false;
	}
	
	uint16_t *depth_data = (uint16_t *)(void *)depthImage.get_buffer();
	uint8* colorBuffer = transformedColorImage.get_buffer();

	k4a_float2_t *xy_table_data = (k4a_float2_t *)(void *)xyLookupTable.get_buffer();
	
	k4a_float2_t p;
	for (unsigned int y = 0; y < height; ++y) {
		p.xy.y = (float)y;
		for (unsigned int x = 0; x < width; ++x) {
			p.xy.x = (float)x;

			int i = x + y * width;

			if (depth_data[i] != 0 && !isnan(xy_table_data[i].xy.x) && !isnan(xy_table_data[i].xy.y))
			{
				tmpPointCloud[i * 4] = -xy_table_data[i].xy.x * (float)depth_data[i] / 10.f;
				tmpPointCloud[i * 4 + 1] = xy_table_data[i].xy.y * (float)depth_data[i] / 10.f;
				tmpPointCloud[i * 4 + 2] = (float)depth_data[i] / 10.f;
			}
			else {
				tmpPointCloud[i * 4] = 0.f;
				tmpPointCloud[i * 4 + 1] = 0.f;
				tmpPointCloud[i * 4 + 2] = 0.f;
			}
		}
	}

	memcpy(tmpColorImage, colorBuffer, width * height * sizeof(uint8_t) * 4);

	depthImage.reset();
	colorImage.reset();

	capture.reset();

	//std::chrono::duration<double> timeDiff = std::chrono::high_resolution_clock::now() - startTime;
	//UE_LOG(LogTemp, Error, TEXT("ReadImage Duration: %f"), timeDiff.count());

	return true;
}

// Called every frame
void AAzureKinect::Tick(float DeltaTime)
{
	Super::Tick(DeltaTime);
}

