// Fill out your copyright notice in the Description page of Project Settings.


#include "PointCloudRenderer.h"

#include "AzureKinect.h"

// Sets default values for this component's properties
UPointCloudRenderer::UPointCloudRenderer()
{
	// Set this component to be initialized when the game starts, and to be ticked every frame.  You can turn these features
	// off to improve performance if you don't need them.
	PrimaryComponentTick.bCanEverTick = true;

	static ConstructorHelpers::FObjectFinder<UNiagaraSystem> NiagaraSystemFinder(TEXT("NiagaraSystem'/Game/PointCloud/PointCloudSystem.PointCloudSystem'"));
	if (NiagaraSystemFinder.Succeeded()) {
		pointCloudRenderer = NiagaraSystemFinder.Object;
	}
}

/**
 * Just a helper method to set a texture for a user variable because the UNiagaraComponent
 * has no direct way to edit a texture variable compared to float, vectors, ...
 */

void SetNiagaraVariableTexture(class UNiagaraComponent* niagara, FString variableName, UTexture* texture) {
	if (!niagara || !texture)
		return;

	FNiagaraUserRedirectionParameterStore& overrideParameters = niagara->GetOverrideParameters();
	FNiagaraVariable niagaraVariable = FNiagaraVariable(FNiagaraTypeDefinition(UNiagaraDataInterfaceTexture::StaticClass()), *variableName);

	UNiagaraDataInterfaceTexture* dataInterface = (UNiagaraDataInterfaceTexture*)overrideParameters.GetDataInterface(niagaraVariable);
	dataInterface->SetTexture(texture);
}


bool UPointCloudRenderer::Init(float* pointCloudArray, uint8_t* colorImageArray, int tWidth, int tHeight, int pDeviceID) {
	deviceID = pDeviceID;

	pointCloudData = pointCloudArray;
	colorImageData = colorImageArray;

	textureWidth = tWidth;
	textureHeight = tHeight;
	pointCount = textureWidth * textureHeight;
	
	region = FUpdateTextureRegion2D(0, 0, 0, 0, textureWidth, textureHeight);

	rendererInstance = UNiagaraFunctionLibrary::SpawnSystemAttached(
		pointCloudRenderer,
		this,
		FName("NiagaraRendererFX" + deviceID),
		FVector(0),
		FRotator(0),
		EAttachLocation::KeepRelativeOffset,
		true
	);

	rendererInstance->SetCastShadow(true);

	// Create dynamic texture for position:
	positionTexture = UTexture2D::CreateTransient(textureWidth, textureHeight, PF_A32B32G32R32F, "PositionData" + deviceID);
	positionTexture->Filter = TF_Nearest;
	positionTexture->UpdateResource();

	// Create dynamic texture for color:
	colorTexture = UTexture2D::CreateTransient(textureWidth, textureHeight, PF_B8G8R8A8, "ColorTexture" + deviceID);
	colorTexture->Filter = TF_Nearest;
	colorTexture->UpdateResource();

	// Set the niagara system user variables:
	SetNiagaraVariableTexture(rendererInstance, "User.PositionTexture", positionTexture);
	SetNiagaraVariableTexture(rendererInstance, "User.ColorTexture", colorTexture);

	rendererInstance->SetVariableInt("User.TextureWidth", textureWidth);
	rendererInstance->SetVariableInt("User.TextureHeight", textureHeight);
	rendererInstance->SetVariableInt("User.PointCount", pointCount);

	FLinearColor monotonicColor(1,0,0,1);

	if (deviceID == 1)
		monotonicColor = FLinearColor(0, 1, 0, 1);
	else if (deviceID == 2)
		monotonicColor = FLinearColor(0, 0, 1, 1);

	rendererInstance->SetVariableLinearColor("User.MonotonicColor", monotonicColor);
#
	positionTexture->UpdateTextureRegions(0, 1, &region, textureWidth * 16, 16, (uint8*)pointCloudData);
	colorTexture->UpdateTextureRegions(0, 1, &region, textureWidth * 4, 4, (uint8*)colorImageData);

	runtime = 0;

	UE_LOG(LogTemp, Warning, TEXT("POINT CLOUD RENDERER is INITIALIZED!!!"));

	bIsInitialized = true;
	return true;
}

// Called when the game starts
void UPointCloudRenderer::BeginPlay()
{
	Super::BeginPlay();

}


// Called every frame
void UPointCloudRenderer::TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
	Super::TickComponent(DeltaTime, TickType, ThisTickFunction);

	runtime += DeltaTime;

	
	// If initialized, update the data every frame:
	if (bIsInitialized) {
		rendererInstance->SetVariableBool("User.UseMonotonicColor", useMonotonicColor);

		positionTexture->UpdateTextureRegions(0, 1, &region, textureWidth * 16, 16, (uint8*)pointCloudData);
		colorTexture->UpdateTextureRegions(0, 1, &region, textureWidth * 4, 4, (uint8*)colorImageData);
	}
}

