// Fill out your copyright notice in the Description page of Project Settings.

#include "RegistrationManager.h"

#include "k4a/k4a.hpp"

#include "IMotionController.h" 	
#include "Features/IModularFeatures.h"

#include "InputCoreTypes.h"

// Sets default values
ARegistrationManager::ARegistrationManager()
{
 	// Set this actor to call Tick() every frame.  You can turn this off to improve performance if you don't need it.
	PrimaryActorTick.bCanEverTick = true;

	RootComponent = CreateDefaultSubobject<USceneComponent>("RootComponent");

	gridMeshComponent = CreateDefaultSubobject<UStaticMeshComponent>("GridMeshComponent");
	gridMeshComponent->SetupAttachment(RootComponent);

	static ConstructorHelpers::FObjectFinder<UStaticMesh> GridMeshFinder(TEXT("StaticMesh'/Game/Geometry/Meshes/Grid.Grid'"));
	if (GridMeshFinder.Succeeded()) {
		gridMeshComponent->SetStaticMesh(GridMeshFinder.Object);
	}
}

// Called when the game starts or when spawned
void ARegistrationManager::BeginPlay()
{
	Super::BeginPlay();

	UE_LOG(LogTemp, Warning, TEXT("Found %i Azure Kinects!") );

	for (unsigned int i = 0; i < k4a::device::get_installed_count(); ++i) {
		AAzureKinect* kinect = GetWorld()->SpawnActor<AAzureKinect>(AAzureKinect::StaticClass());
		kinect->Init(i);
		kinects.Add(kinect);
	}

	thread = new FRegistrationThread((ARegistrationManager*)this);
}

void ARegistrationManager::EndPlay(const EEndPlayReason::Type EndPlayReason) {
	if (thread)
		thread->Terminate();

	for (AAzureKinect* kinect : kinects) {
		if (kinect->IsInitialized()) {
			kinect->Shutdown();
		}
	}
}

void ARegistrationManager::UpdateMotionControllerPositions(FVector left, FVector right) {
	leftControllerPosition = left;
	rightControllerPosition = right;
	controllerPositionsAvailable = true;
}


void ARegistrationManager::RegisterMultipleKinects() {
	// First, cache point clouds so we use point clouds with similar time stamps
	// (not perfect synchronized, but hopefully good enough):
	TArray<float*> cachedPointCloudData;
	{
		for (AAzureKinect* kinect : kinects) {
			int size = kinect->getWidth() * kinect->getHeight();

			float* cachedPC = new float[size * 3];

			float* kinectPC = kinect->currentTmpPointCloudData();
			// Scale from centimeters to meters:
			for (int i = 0; i < size; ++i) {
				cachedPC[i * 3] = kinectPC[i * 4] / 100.f;
				cachedPC[i * 3 + 1] = kinectPC[i * 4 + 1] / 100.f;
				cachedPC[i * 3 + 2] = kinectPC[i * 4 + 2] / 100.f;
			}

			cachedPointCloudData.Add(cachedPC);
		}
	}

	TArray<GridDetectionResult> gridDetectionResults;

	// Find grid in all kinect images:
	for (int i = 0; i < kinects.Num(); ++i) {
		// Create point cloud image definition:
		PointCloudImage image(cachedPointCloudData[i], kinects[i]->getWidth(), kinects[i]->getHeight());

		// Detect grid:
		GridDetectionResult gdr = GridDetector::detectGrid(image);
		gridDetectionResults.Add(gdr);
		UE_LOG(LogTemp, Warning, TEXT("GridDetection for Kinect %i: %s"), i, gdr.gridFound ? TEXT("FOUND") : TEXT("-"));
	}

	// Match hole points of all gridDetectionResults:
	TArray<TArray<FVector>> correspondencesForRoughRegistration = CorrespondencePointMatcher::matchGridCenterAndXAxis(gridDetectionResults);

	// If the number of point clouds do not fit, recreate a fitting point cloud registrator:
	if (multiplePCRegistrator.getPointCloudCount() != kinects.Num()) {
		roughPCRegistrator = MultiplePointCloudsRegistrator(kinects.Num());
		multiplePCRegistrator = MultiplePointCloudsRegistrator(kinects.Num());
	}

	roughPCRegistrator.addCorrespondencePoints(correspondencesForRoughRegistration);

	TArray<TArray<FVector>> allCorrespondences = CorrespondencePointMatcher::matchAllGridHoles_ByRoughRegistration(
		roughPCRegistrator.registerInto(0, correspondenceRejectionKinectsThreshold),
		gridDetectionResults,
		0
	);

	// Add the points detected by the grid:
	multiplePCRegistrator.addCorrespondencePoints(allCorrespondences);

	// Get results for registrate all the point clouds into the first point cloud:
	currentRegistration = multiplePCRegistrator.registerInto(0, correspondenceRejectionKinectsThreshold);

	// DELETE CACHED POINT CLOUD DATA;
	for (float* v : cachedPointCloudData) {
		delete[] v;
	}
}

void ARegistrationManager::RegisterKinectsIntoHMD() {
	if (controllerPositionsAvailable) {
		int size = kinects[0]->getWidth() * kinects[0]->getHeight();

		float* scaledPC = new float[size * 3];

		float* kinectPC = kinects[0]->currentTmpPointCloudData();
		// Scale from centimeters to meters:
		for (int i = 0; i < size; ++i) {
			scaledPC[i * 3] = kinectPC[i * 4] / 100.f;
			scaledPC[i * 3 + 1] = kinectPC[i * 4 + 1] / 100.f;
			scaledPC[i * 3 + 2] = kinectPC[i * 4 + 2] / 100.f;
		}

		PointCloudImage image(scaledPC, kinects[0]->getWidth(), kinects[0]->getHeight());
		GridDetectionResult gdr = GridDetector::detectGrid(image);

		if(gdr.gridFound)
			debugDrawGridTransform = FTransform(gdr.axis1, gdr.axis2, gdr.normal, gdr.gridCenter * 100);

		delete[] scaledPC;

		MotionControllerAssumptions assumptions;
		assumptions.leftOffsetFromGridMid = leftControllerOffset;
		assumptions.rightOffsetFromGridMid = rightControllerOffset;
		assumptions.swapControllers = swapControllers;

		MotionControllerResult mcr = CorrespondencePointMatcher::findMotionControllerRelativeToGrid(gdr, assumptions);

		if (!mcr.found)
			return;

		debugDrawLeftControllerCS = mcr.leftControllerPosition * 100.f;
		debugDrawRightControllerCS = mcr.rightControllerPosition * 100.f;

		if (pcWithHMDRegistrator.getPointCloudCount() != 2) {
			pcWithHMDRegistrator = MultiplePointCloudsRegistrator(2);
		}

		FVector l = leftControllerPosition / 100.f;
		FVector r = rightControllerPosition / 100.f;
		controllerPositionsAvailable = false;
		TArray<TArray<FVector>> correspondencePoints = CorrespondencePointMatcher::matchControllerAndGridDetectionResult(l, r, mcr);
		currentRegistration = multiplePCRegistrator.registerInto(0);

		pcWithHMDRegistrator.addCorrespondencePoints(correspondencePoints);

		TArray<FTransform> registrationResult = pcWithHMDRegistrator.registerInto(0, correspondenceRejectionToWorldThreshold);


		if (registrationResult.Num() > 0) {
			kinect0IntoWorldRegistration = registrationResult[1];

			meanKinect0ToWorldError = pcWithHMDRegistrator.getMeanError(1);
			UE_LOG(LogTemp, Warning, TEXT("Motion Controller, Left: %f, %f, %f | Right: %f, %f, %f"), l.X, l.Y, l.Z, r.X, r.Y, r.Z);

			correspondenceCountVRTotal = pcWithHMDRegistrator.getLastTotalCorrespondencePointSize(0);
			correspondenceCountVRUsed = pcWithHMDRegistrator.getLastUsedCorrespondencePointSize(0);
		}
	}
	else {
		UE_LOG(LogTemp, Warning, TEXT("Motion Controller Poisitions were not available this frame."))
	}

}

void ARegistrationManager::RegisterOnCurrent() {
	if (shouldRegisterMultipleKinects) {
		RegisterMultipleKinects();
	}

	if (shouldRegisterKinectIntoHMD) {
		RegisterKinectsIntoHMD();
	}
}

// Called every frame
void ARegistrationManager::Tick(float DeltaTime)
{
	Super::Tick(DeltaTime);

	// Transform the kinects so that their point clouds will be registrated into the point cloud of the first kinect:
	if (kinects.Num() >= 1 && !FTransform::Identity.Equals(kinect0IntoWorldRegistration)) {
		kinects[0]->SetActorTransform(kinect0IntoWorldRegistration);
		kinects[0]->gridMeshComponent->SetRelativeTransform(debugDrawGridTransform);
		kinects[0]->estimatedLeftController->SetRelativeLocation(debugDrawLeftControllerCS);
		kinects[0]->estimatedRightController->SetRelativeLocation(debugDrawRightControllerCS);
	}

	if (currentRegistration.Num() >= kinects.Num()) {
		for (int i = 1; i < kinects.Num(); ++i) {
			kinects[i]->SetActorTransform(currentRegistration[i] * kinects[0]->GetActorTransform());
		}
	}
}


